import base64
from io import BytesIO
from enum import Enum
from pprint import pprint

from google.genai import types
from llm_bridge.type.chat_response import Citation, ChatResponse


class PrintingStatus(Enum):
    Start = "start"
    Thought = "thought"
    Response = "response"


class GeminiResponseHandler:
    def __init__(self):
        self.printing_status = PrintingStatus.Start

    def process_gemini_response(
            self, response: types.GenerateContentResponse
    ) -> ChatResponse:
        text = ""
        display = None
        image_base64 = None
        citations = extract_citations(response)

        if response.candidates[0].content.parts:
            for part in response.candidates[0].content.parts:
                if part.text:
                    if part.thought and self.printing_status == PrintingStatus.Start:
                        text += "# Model Thought:\n\n"
                        self.printing_status = PrintingStatus.Thought
                    elif not part.thought and self.printing_status == PrintingStatus.Thought:
                        text += f"\n\n# Model Response:\n\n"
                        self.printing_status = PrintingStatus.Response
                    text += part.text
                elif part.inline_data:
                    image_base64 = base64.b64encode(part.inline_data.data).decode('utf-8')

        if grounding_metadata := response.candidates[0].grounding_metadata:
            if search_entry_point := grounding_metadata.search_entry_point:
                display = search_entry_point.rendered_content
            if grounding_metadata.grounding_chunks:
                text += "\n\n# Grounding Sources:\n"
                for i, chunk in enumerate(grounding_metadata.grounding_chunks, start=1):
                    if chunk.web:
                        text += f"{i}. [{chunk.web.title}]({chunk.web.uri})\n"

        return ChatResponse(text=text, image=image_base64, display=display, citations=citations)


def extract_citations(response: types.GenerateContentResponse) -> list[Citation]:
    citations = []
    if grounding_metadata := response.candidates[0].grounding_metadata:
        if grounding_supports := grounding_metadata.grounding_supports:
            for grounding_support in grounding_supports:
                citation_indices = [index + 1 for index in grounding_support.grounding_chunk_indices]
                citation_text = grounding_support.segment.text
                citations.append(Citation(text=citation_text, indices=citation_indices))
    return citations if citations else None


from fastapi import HTTPException

from llm_bridge.client.chat_client import ChatClient
from llm_bridge.logic.chat_generate.model_client_factory.claude_client_factory import create_claude_client
from llm_bridge.logic.chat_generate.model_client_factory.gemini_client_factory import create_gemini_client
from llm_bridge.logic.chat_generate.model_client_factory.gpt_client_factory import create_gpt_client
from llm_bridge.type.message import Message


async def create_chat_client(
        messages: list[Message],
        model: str,
        api_type: str,
        temperature: float,
        stream: bool,
        api_keys: dict
) -> ChatClient:
    if api_type == 'OpenAI':
        return await create_gpt_client(
            messages=messages,
            model=model,
            api_type=api_type,
            temperature=temperature,
            stream=stream,
            api_keys={"OPENAI_API_KEY": api_keys["OPENAI_API_KEY"]}
        )
    elif api_type == 'OpenAI-Azure':
        return await create_gpt_client(
            messages=messages,
            model=model,
            api_type=api_type,
            temperature=temperature,
            stream=stream,
            api_keys={
                "AZURE_API_KEY": api_keys["AZURE_API_KEY"],
                "AZURE_API_BASE": api_keys["AZURE_API_BASE"]
            }
        )
    elif api_type == 'OpenAI-GitHub':
        return await create_gpt_client(
            messages=messages,
            model=model,
            api_type=api_type,
            temperature=temperature,
            stream=stream,
            api_keys={"GITHUB_API_KEY": api_keys["GITHUB_API_KEY"]}
        )
    elif api_type == 'Grok':
        return await create_gpt_client(
            messages=messages,
            model=model,
            api_type=api_type,
            temperature=temperature,
            stream=stream,
            api_keys={"XAI_API_KEY": api_keys["XAI_API_KEY"]}
        )
    elif api_type == 'Gemini':
        return await create_gemini_client(
            messages=messages,
            model=model,
            temperature=temperature,
            stream=stream,
            api_key=api_keys["GOOGLE_AI_STUDIO_API_KEY"]
        )
    elif api_type == 'Claude':
        return await create_claude_client(
            messages=messages,
            model=model,
            temperature=temperature,
            stream=stream,
            api_key=api_keys["ANTHROPIC_API_KEY"]
        )
    else:
        raise HTTPException(status_code=400, detail="Invalid API type")



import pytest

from llm_bridge.logic.message_preprocess.message_preprocessor import extract_system_messages
from llm_bridge.type.message import Message, Role, Content, ContentType


@pytest.fixture
def sample_messages():
    return [
        Message(role=Role.System, contents=[
            Content(type=ContentType.Text, data="You are a helpful assistant.")
        ]),
        Message(role=Role.User, contents=[
            Content(type=ContentType.Text, data="Hello")
        ])
    ]

def test_extract_system_messages(sample_messages):
    extracted_text = extract_system_messages(sample_messages)

    assert extracted_text == "You are a helpful assistant.\n"

    assert len(sample_messages) == 1
    assert sample_messages[0].role == Role.User
    assert sample_messages[0].contents[0].type == ContentType.Text
    assert sample_messages[0].contents[0].data == "Hello"



import logging
from io import BytesIO

import httpcore
import httpx
from fastapi import HTTPException
from tenacity import retry, stop_after_attempt, wait_exponential, retry_if_exception_type


@retry(
    stop=stop_after_attempt(3),
    wait=wait_exponential(multiplier=1, min=1, max=4),
    retry=retry_if_exception_type((httpx.ConnectError, httpcore.ConnectError)),
    reraise=True,
)
async def fetch_file_data(file_url: str) -> tuple[BytesIO, str]:
    async with httpx.AsyncClient() as client:
        try:
            response = await client.get(file_url)
        except httpx.ConnectError as e:
            logging.exception(f"httpx.ConnectError while fetching {file_url}: {e}")
            raise
        except httpcore.ConnectError as e:
            logging.exception(f"httpcore.ConnectError while fetching {file_url}: {e}")
            raise
        except Exception as e:
            logging.exception(f"Unknown error while fetching {file_url}: {e}")
            raise e

        if response.status_code != 200:
            status_code = response.status_code
            text = response.text
            logging.error(f"Error {status_code}: {text}")
            raise HTTPException(status_code=status_code, detail=text)

        content_type = response.headers.get("Content-Type", "")
        return BytesIO(response.content), content_type



from dataclasses import dataclass
from enum import Enum


class Role(Enum):
    User = "user"
    Assistant = "assistant"
    System = "system"


class ContentType(Enum):
    Text = "text"
    File = "file"


@dataclass
class Content:
    type: ContentType
    data: str # text or file path


@dataclass
class Message:
    role: Role
    contents: list[Content]



from llm_bridge.logic.chat_generate.model_message_converter.claude_message_converter import convert_message_to_claude
from llm_bridge.logic.chat_generate.model_message_converter.gpt_message_converter import convert_message_to_gpt
from llm_bridge.logic.chat_generate.model_message_converter.gemini_message_converter import convert_message_to_gemini
from llm_bridge.type.model_message.claude_message import ClaudeMessage
from llm_bridge.type.model_message.gemini_message import GeminiMessage
from llm_bridge.type.model_message.gpt_message import GptMessage
from llm_bridge.type.message import Message


async def convert_messages_to_gemini(messages: list[Message]) -> list[GeminiMessage]:
    return [await convert_message_to_gemini(message) for message in messages]


async def convert_messages_to_gpt(messages: list[Message]) -> list[GptMessage]:
    return [await convert_message_to_gpt(message) for message in messages]


async def convert_messages_to_claude(messages: list[Message]) -> list[ClaudeMessage]:
    return [await convert_message_to_claude(message) for message in messages]



import openai.lib.azure

from llm_bridge.client.implementations.gpt.non_stream_gpt_client import NonStreamGPTClient
from llm_bridge.client.implementations.gpt.stream_gpt_client import StreamGPTClient
from llm_bridge.logic.chat_generate.model_message_converter.model_message_converter import convert_messages_to_gpt
from llm_bridge.type.message import Message


async def create_gpt_client(
        messages: list[Message],
        model: str,
        api_type: str,
        temperature: float,
        stream: bool,
        api_keys: dict
):
    client = None
    if api_type == "OpenAI":
        client = openai.AsyncOpenAI(
            api_key=api_keys["OPENAI_API_KEY"],
        )
    elif api_type == "Azure":
        client = openai.lib.azure.AsyncAzureOpenAI(
            api_version="2024-02-01",
            azure_endpoint=api_keys["AZURE_API_BASE"],
            api_key=api_keys["AZURE_API_KEY"],
        )
    elif api_type == "GitHub":
        client = openai.AsyncOpenAI(
            base_url="https://models.inference.ai.azure.com",
            api_key=api_keys["GITHUB_API_KEY"],
        )
    elif api_type == "Grok":
        client = openai.AsyncOpenAI(
            base_url="https://api.x.ai/v1",
            api_key=api_keys["XAI_API_KEY"],
        )

    gpt_messages = await convert_messages_to_gpt(messages)

    if stream:
        return StreamGPTClient(
            model=model,
            messages=gpt_messages,
            temperature=temperature,
            api_type=api_type,
            client=client,
        )
    else:
        return NonStreamGPTClient(
            model=model,
            messages=gpt_messages,
            temperature=temperature,
            api_type=api_type,
            client=client,
        )


from typing import AsyncGenerator

import openai.lib.azure

from llm_bridge.client.chat_client import ChatClient
from llm_bridge.type.chat_response import ChatResponse
from llm_bridge.type.model_message.gpt_message import GptMessage


class GPTClient(ChatClient):
    def __init__(
            self,
            model: str,
            messages: list[GptMessage],
            temperature: float,
            api_type: str,
            client: openai.AsyncOpenAI | openai.lib.azure.AsyncAzureOpenAI,
    ):
        self.model = model
        self.messages = messages
        self.temperature = temperature
        self.api_type = api_type
        self.client = client

    async def generate_non_stream_response(self) -> ChatResponse:
        raise NotImplementedError
    async def generate_stream_response(self) -> AsyncGenerator[ChatResponse, None]:
        raise NotImplementedError
